/*: TAVLREBL.C	 TAVL Tree library    Version 3.01   6-feb-93
 *: Functions:	tavl_rebalance, minOf, maxOf, parentOf
 *: Purpose: Private functions for internal use of TAVL library
 *	     Rebalance the TAVLtree "a", which is unbalanced by at most
 *	     one leaf.	This function may be called by either "tavl_insert"
 *	     or "tavl_delete"; Rebalance is considered to be private to the
 *	     TAVL routines, its prototype is in the header "tavlpriv.h",
 *	     whereas user routine prototypes are in "tavltree.h".
 *	     Assumes balance factors & threads are correct; Returns pointer
 *	     to root of balanced tree; threads & balance factors have been
 *	     corrected if necessary.  If the height of the rebalanced subtree
 *	     is less than the height of the original subtree "a", *deltaht
 *	     is set to 1, otherwise *deltaht is set to 0.  If "tavl_rebalance"
 *	     is called from "tavl_insert", *deltaht will always be set to 1 -
 *	     just by the nature of the algorithm - however, when called from
 *	     "tavl_delete" the height of the new subtree is not necessarily
 *	     less than the original subtree "a"'s height, instead, rebalance
 *	     may simply have moved a node up. Tavl_delete needs to know which
 *	     case occurred.
 *
 *	     Functions "minOf, maxOf, parentOf" are also defined in this
 *	     file, and are used exclusively by "tavl_delete" and "tavl_insert".
 *
 *  Author:  Bert C. Hughes
 *
 * !!NOTE!!  This module must be linked into any program which uses
 *	     the TAVL library functions "tavl_insert" or "tavl_delete"!
 *	     This module is strictly for the use of those functions, and
 *	     is NOT intended to be a "user", or TAVL library, function.
 *
 *  Released to the PUBLIC DOMAIN
 *			    Bert C. Hughes
 *			    200 N.Saratoga
 *			    St.Paul, MN 55104
 *			    Compuserve 71211,577
 */

   /* Debugging "assert"s are in the code to test integrity of the    */
   /* tree. All assertions should be removed from production versions */
   /* of the library by compiling the library with NDEBUG defined.    */
   /* See the header "assert.h".*/

#include <stdlib.h>	  /* for prototype or macro "abs()" */
#include <assert.h>
#include "tavlpriv.h"

#pragma warn -rvl

TAVL_nodeptr rebalance_tavl(TAVL_nodeptr a, char *deltaht)
/* rebalance subtree "a", connect to "a"'s parent, return pointer
 * to root of newly rebalanced subtree */
{
    TAVL_nodeptr b,c,p;
		    /*	definition(tree-height(X)) : the maximum    */
		    /*	    path length from node X to a leaf node. */
    *deltaht = 0;   /*	*deltaht is set to 1 if and only if	    */
		    /*	    tree-height(rebalance()) < tree-height(a)*/

    if (Is_Head(a)	    /* Never rebalance the head node! */
     || abs(a->bf) < 2)     /* subtree "a" is balanced - nothing more to do */
	return a;

    p = parentOf(a);

    if (a->bf == LEFT+LEFT) {
	b = a->Lptr;
	if (b->bf != RIGHT) {	/* LL rotation */

	    if (a->rchild)
	       { p->Rptr = b;  b->rchild = 1; }
	    else
	       { p->Lptr = b;  a->rchild = 1; }

	    if (RTHREAD(b)) {	    /* b->Rptr is a thread to "a" */
		assert(b->Rptr == a);
		a->Lbit = THREAD;   /* change from link to thread */
		b->Rbit = LINK;     /* change thread to link */
	    }
	    else {
		a->Lptr = b->Rptr;
		b->Rptr = a;
		a->Lptr->rchild = 0;
	    }

	    *deltaht = b->bf ? 1 : 0;

	 /* a->bf = - (b->bf += RIGHT);  See HISTORY.TXT:
	    This is a workaround for slightly bugged compilers...
	  */
	    b->bf += RIGHT;   /* the workaround */
	    a->bf = - b->bf;

	    return b;
	}
	else {			/* LR rotation */
	    *deltaht = 1;

	    c = b->Rptr;

	    if (a->rchild)
	       p->Rptr = c;
	    else
	       { p->Lptr = c;  a->rchild = 1;  c->rchild = 0; }

	    if (LTHREAD(c)) {
		assert(c->Lptr == b);
		c->Lbit = LINK;
		b->Rbit = THREAD;
	    }
	    else {
		b->Rptr = c->Lptr;
		b->Rptr->rchild = 1;
		c->Lptr = b;
	    }

	    if (RTHREAD(c)) {
		assert(c->Rptr == a);
		c->Rbit = LINK;
		a->Lptr = c;
		a->Lbit = THREAD;
	    }
	    else {
		a->Lptr = c->Rptr;
		a->Lptr->rchild = 0;
		c->Rptr = a;
	    }

	    switch (c->bf) {
		case LEFT:  b->bf = 0;
			    a->bf = RIGHT;
			    break;

		case RIGHT: b->bf = LEFT;
			    a->bf = 0;
			    break;

		case 0:     b->bf = 0;
			    a->bf = 0;
	    }

	    c->bf = 0;

	    return c;
	}
    }
    else if (a->bf == RIGHT+RIGHT) {
	b = a->Rptr;
	if (b->bf != LEFT) {	/* RR rotation */

	    if (a->rchild)
	       { p->Rptr = b;  a->rchild = 0; }
	    else
	       { p->Lptr = b;  b->rchild = 0; }

	    if (LTHREAD(b)) {	    /* b->Lptr is a thread to "a" */
		assert(b->Lptr == a);
		a->Rbit = THREAD;   /* change from link to thread */
		b->Lbit = LINK;     /* change thread to link */
	    }
	    else {
		a->Rptr = b->Lptr;
		a->Rptr->rchild = 1;
		b->Lptr = a;
	    }

	    *deltaht = b->bf ? 1 : 0;

	   /* a->bf = - (b->bf += LEFT);  See HISTORY.TXT:
	    This is a workaround for slightly bugged compilers...
	  */
	    b->bf += LEFT;    /* the workaround */
	    a->bf = - b->bf;

	    return b;
	}
	else {			/* RL rotation */
	    *deltaht = 1;

	    c = b->Lptr;

	    if (a->rchild)
	       { p->Rptr = c;  c->rchild = 1;  a->rchild = 0; }
	    else
	       p->Lptr = c;

	    if (RTHREAD(c)) {
		assert(c->Rptr == b);
		c->Rbit = LINK;
		b->Lbit = THREAD;
	    }
	    else {
		b->Lptr = c->Rptr;
		b->Lptr->rchild = 0;
		c->Rptr = b;
	    }

	    if (LTHREAD(c)) {
		assert(c->Lptr == a);
		c->Lbit = LINK;
		a->Rptr = c;
		a->Rbit = THREAD;
	    }
	    else {
		a->Rptr = c->Lptr;
		a->Rptr->rchild = 1;
		c->Lptr = a;
	    }

	    switch (c->bf) {
		case RIGHT: b->bf = 0;
			    a->bf = LEFT;
			    break;

		case LEFT:  b->bf = RIGHT;
			    a->bf = 0;
			    break;

		case 0:     b->bf = 0;
			    a->bf = 0;
	    }

	    c->bf = 0;

	    return c;
	}
    }

    assert(0); /* can't happen */

}/* end rebalance */


/* return pointer to minimum node of subtree t */
TAVL_nodeptr minOf(TAVL_nodeptr t)
{
    while (LLINK(t)) t = t->Lptr;
    return t;
}

/* return pointer to maximum node of subtree t */
TAVL_nodeptr maxOf(TAVL_nodeptr t)
{
     while (RLINK(t)) t = t->Rptr;
     return t;
}

/* return pointer to parent of subtree t */
TAVL_nodeptr parentOf(TAVL_nodeptr t)
{
     return (t->rchild ? minOf(t)->Lptr : maxOf(t)->Rptr);
}

